#ifndef _DIDERO_PC_H
#define	_DIDERO_PC_H
/* Dynamic Link Library related definitions*/
//#include "C:/Programming_Libraries/dlfcn/include/dlfcn.h"
//#include "dlfcn.h"

/* Interface definitions */
#include "Client_Interface.h"
#include "Server_Interface.h"
#include "Server_Interface_2.h"
#include "Middleware_Interface.h"

/* SAN project files */
//#include "../Atomic/TransactionSAN_1/TransactionSAN_1SAN.h"
#include "Cpp/BaseClasses/SAN/Place.h"
#include "Cpp/Simulator/SimulatorClass.h"

/* relevant standard c++ and stl header files*/
#include <vector>
#include <fstream>
#include <sstream>

/* For some math definitions, including the smallest "double" DBL_MIN*/
#include <float.h>

/* Definition of UDF class that implements a Client for sending Trace data to listening Server */
#include "Inter_Proc_Comm_Client.h"

extern void Initialize_Clients_and_Servers();
extern std::string Begin_Transaction(int ID);
extern bool Next_Operation_Method(int ID);
extern string Exec_Trans_Method(int ID, int Server_ID);
extern bool Check_Blocked(int ID, int Server_activated);
extern string DivSQL_Results(int MW_ID, int Server_ID, int Server_1_Waiting, int Server_2_Waiting, int Wait_for_Slower_Server,  int Wait_for_faster_Server);
extern string DivSQL_Requests(int MW_ID, int Slower_Server);
extern double Duration(int Trans_ID, int Server_ID, double current_time);
extern bool Reac_Pred(int Trans_ID, int All_Serv_Trans, int All_CommitSuccess, int Commit_This_Trans);
extern double DivREP_Requests_Duration(int commit, int MW_ID);
extern double DivREP_Results_Duration(int MW_ID);
extern int Init_Emp_Dists(string PathToFileWithDataLocations);
extern double sample(map<double, double>* dist, bool new_quantile, bool reset_previous_quantile, int Trans_ID, int Server_ID, double current_sim_time, bool reactivate);
extern double estimate_Percentile(map<double, double>* dist, double percentile);
extern bool Server_Reac_Pred(int Trans_ID, int Start_Oper, int End_Oper);
extern bool areEqualRel(double a, double b, double epsilon);

extern short GV_Enable_Client_View_Trace;
extern short GV_Enable_Server_1_View_Trace;
extern short GV_Enable_Server_2_View_Trace;
extern short GV_Number_of_Warehouses;
extern short GV_Num_Dist_per_Warehouses;
extern int GV_Cust_per_District;
extern short GV_Number_of_Clients_and_Trans;
extern short GV_debug_flag;
extern int GV_Efficient_Oper_Capacity_Server_1;
extern int GV_Efficient_Oper_Capacity_Server_2;
extern double GV_ReacModel_ExpScaling;
extern double GV_ExpScaling_LowerBound;
extern double GV_ReacModel_PolyPower;
extern int GV_Experiment_Runs;
extern float GV_Oper_Dur_Raw_Data_Scaling;
extern bool GV_Server_1_enabled;
extern bool GV_Server_2_enabled;
extern bool GV_Step_Inverse_Transform;
extern bool GV_TPCC;
extern bool GV_Couple_Server_Trans_Durations;
extern bool GV_IntraTransaction_Dependence;
extern bool GV_IntraTransDepBurstiness;
extern bool GV_DefaultResidualTimes;
extern double GV_Server_1_Prob_of_IntraTransDependence;
extern double GV_Server_2_Prob_of_IntraTransDependence;

/* Defined Mobius Global variables constants for different distributions */
/* Discrete Distribution of TPCC transaction profile occurrence */
extern double Prob_of_NewOrder;
extern double Prob_of_Payment;
extern double Prob_of_Orderstatus;
extern double Prob_of_Delivery;
extern double Prob_of_Stocklevel;

/* Order Status */
/* Choose customers according to ID with 40% probability */
extern double Choose_Cust_using_ID;

/* Choose local warehouse with 90% probability */
extern double Prob_of_Local_Orderlines;

/* Payment */
/* Choose Customer using ID or lastname */
extern double Prob_of_Cust_ID;

/* Choose a bad customer with 1% probability */
extern double Prob_of_Bad_Cust;                


extern int Experiment_Runs_Count;

extern vector<int> ActionFiringCounts;  /* The data structure holds pairs of counts which, together, indicate how many times data is sent to the Trace logger */

extern vector<bool> Exec_Oper_on_both_Servers;         /* Array of booleans to indicate that an operation 
                                                 * has been carried out on both servers. 
                                                 */
extern vector<Client_Interface*> ClientObjects;        /* Array of pointers to Client Object interface*/
extern Server_Interface* Server_1;        /* Pointer to Server Object interface*/
extern Server_Interface* Server_2;        /* Pointer to Server Object interface*/
extern Middleware_Interface* MW_ptr;      /* Pointer to Middleware Object Interface */      
extern map< int, Place* > Unblocked_place_ptrs;        /* Pointers to places that unblock blocked Transitions */
extern map< int, Place* > Blocked_place_ptrs;        /* Pointers to places that indicate the blocked status of each Transitions */
extern vector<string> Oper_results;             /* Array of strings to store the result of the faster server */
extern map<int, map< int, map< int , string> > > Trans_Oper_Replica_Results;       /* Stores the results of operations executed by the 
Client_ID  Operation_ID   Server_ID                                            replicas. These results are compared by the Middlewware 
                                                                             * for possible inconsistencies. 
                                                                             */

extern map<int, map<int, vector<string> > > Oper_Data_for_Replicas;   /*Array to store the data that defines each operation in a Transaction */
/* Client_ID Operation_ID   Data */

/* Data structure that keeps count of which operation is currently being executed by a given replica, with respect to a given client */
extern map<int, map<int, int>  > Server_oper_index;
extern map< int, map<int, map<double, double> > > Emp_Distributions;       /* Data structure to store Empirical Distributions */

/* Data structure to hold the generated sojourn-time values for operations on each Server */
extern map< int, map<int, double> > Sampled_SojournTimes;
/* Data structure to keep a count of the number of operations currently running on each Server */
extern map< int, int > Current_Num_of_Oper;
/* Data structure to hold, for each operation on each server, the last time a sojourn-time for this operation
 * was generated, except in those cases where an operation is new, in which case the associated time value
 * should be zero (indicating the operation is having its first sojourn-time computed ).
 */
extern map< int, map<int, double> > PreviousEventTime;

/******** Data structure to indicate when a "Begin" has occurred on the respective Server ********/
extern map<int, map<int, bool > > Begin_Status;
/* Server_Trans_ID_boolean */

extern map<int,int> Unbounded_Trans_ID;        /* Data structure to indicate how many transactions have been executed so far for a given Client */

extern Inter_Proc_Comm_Client* Trace_data_Client;      /* Pointer to object that implements Client behaviour in a socket
                                                 * based solution for sending Trace data to a listening server 
                                                 */
#endif	/* _DIDERO_PC_H */
