#include "DIDERO_PC.h"
#include <map>
#include <vector>

using namespace std;

/* The following declarations are here simply for convenience; it is already made in DIDERO_PC.h */
extern vector<Client_Interface*> ClientObjects;        /* Array of pointers to Client Object interface*/
extern map< int, map<int, map< double, double> > > Emp_Distributions;       /* Data structure to store Empirical Distributions */ 

/**** Mobius Global Variables *****/
extern double GV_ReacModel_ExpScaling;
extern double GV_ReacModel_PolyPower;

/******** Data structure to indicate when a "Begin" has occurred on the respective Server ********/
map<int, map<int, bool > > Begin_Status;
/* Server_Trans_ID_boolean */

/* Data-structure used to store estimated sojourn-times for each operation on each server */
extern map< int, map< int, double > > previous_sampled_sojtimes;
/* Client_ID : Server_ID : duration */

/************ Header files for Boost Distributions *************/

/* I think the following header contains the typedef for minstd_rand */
#include <boost/random/linear_congruential.hpp>

#include "boost/random.hpp"

/* I think this header file is for the "definition" of the "variate_generator" data-type  */
#include <boost/random/variate_generator.hpp>

#include <boost/random/exponential_distribution.hpp>

/* KS comment (17-08-2010: )The following appears to be a typedef of a "data-type" for boost's random number seed generators */
typedef boost::minstd_rand base_generator_type;

/* Initialize random number generation functions */
static base_generator_type generator(42u);

/* Define an exponentially distributed random variable with parameter 100. */
boost::exponential_distribution<> ExpDist(100);
boost::variate_generator<base_generator_type&, boost::exponential_distribution<> > SojournTimeValue(generator, ExpDist);

/* Data structure to hold the generated sojourn-time values for operations on each Server */
map< int, map<int, double> > Sampled_SojournTimes;

/* Data structure to keep a count of the number of operations currently running on each Server */
map< int, int > Current_Num_of_Oper;

/* Data structure to hold, for each operation on each server, the last time a sojourn-time for this operation
 * was generated, except in those cases where an operation is new, in which case the associated time value
 * should be zero (indicating the operation is having its first sojourn-time computed ).
 */
map< int, map<int, double> > PreviousEventTime;

/* Declaration for method defined below */
double Residual_Sojourn_Time(double rv, int Server_ID, int Trans_ID, double TimeSinceLastEvent);

/* Declaration for method to compare doubles */
bool areEqualRel(double a, double b, double epsilon);

/* Declaration for Method that samples Empirical Distribution */
double Sample_Dist(int Trans_ID, int Server_ID, int Oper_case, double Time_Since_Last_Event, bool quantile_flag, double current_sim_time, bool my_reactivate_boolean);

/* Declaration of method for sending trace information to listening "socket" server */
void print_trace_method(bool print_trace, int Server_ID, int Trans_ID, double current_time, std::string operation, int Operation_ID);

double Duration(int Trans_ID, int Server_ID, double current_time)
{
    double duration;
    bool new_quantile = false;
    bool print_trace = false;
    bool my_reactivate_boolean = false;   
    
    double Time_Since_Last_Event = 0;
    
    //std::numeric_limits<double>::epsilon()
    if(GV_debug_flag==23)
    {
        std::cerr << "\n\n\n\nDurations.cpp:: The number of operations currently executing is " << Current_Num_of_Oper[Server_ID] << " , on Server " << Server_ID << " !!!!"<< std::endl;
    }        
        
    /* Check to see if the time of the the last event has been stored as zero, indicating that this is a new transaction operation  */
    //if((PreviousEventTime[Server_ID])[Trans_ID]!=0)
    if(!areEqualRel((PreviousEventTime[Server_ID])[Trans_ID], 0, std::numeric_limits<double>::epsilon()))
    {
        /* This is a reactivation */
        my_reactivate_boolean = true;
        Time_Since_Last_Event = current_time - (PreviousEventTime[Server_ID])[Trans_ID];

        //if(GV_debug_flag==23)
        //if(Trans_ID==5)
        if(GV_debug_flag==23)
        {
            std::cerr << "Durations.cpp:: The values of (Sampled_SojournTimes[Server_ID])[Trans_ID], current_time and (PreviousEventTime[Server_ID])[Trans_ID] are " << (Sampled_SojournTimes[Server_ID])[Trans_ID] << " , " << current_time << " and "<< (PreviousEventTime[Server_ID])[Trans_ID] << " , on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " . This is a REACTIVATION!!!!"<< std::endl;
        }        
    }else
    {
        /* The sojourn-time for this operation is being computed for the first time.
         * 
         *  
         */
        Time_Since_Last_Event = 0;
        //if(GV_debug_flag==14)
        if(GV_debug_flag==23)
        {
            std::cerr << "Durations.cpp:: The values of (Sampled_SojournTimes[Server_ID])[Trans_ID], current_time and (PreviousEventTime[Server_ID])[Trans_ID] are " << (Sampled_SojournTimes[Server_ID])[Trans_ID] <<" , " << current_time << " and "<< (PreviousEventTime[Server_ID])[Trans_ID] << " , on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " . This is a NEW operation!!!!"<< std::endl;
        }
        
        /* print trace data on this occasion */
        print_trace = true;
    }

    /* The overall structure of the code below goes through the following possibilities:
     * Is this a "begin"? Ok, use the respective empirical distribution;
     * No. Is this a "Commit"? Ok, use the respective empirical distribution;
     * No. Is this an Abort? Ok, use the respective empirical distribution;
     * Then this is a normal operation. Ok, use the respective empirical distribution; 
     */
    
    /* If the "Begin_Status" flag is false and only one operation has been requested of the Server to perform
     * --- that is, no operation has been performed yet -- then this is a begin operation  
     */

   if(GV_debug_flag==23)
   {
       std::cerr << "Durations.cpp:: The number of operations performed so far is " << (Server_oper_index[Trans_ID])[Server_ID] << std::endl;
       std::cerr << "Durations.cpp:: The value of ((Begin_Status[Server_ID])[Trans_ID]) is  " << (Begin_Status[Server_ID])[Trans_ID] << std::endl;
       std::cerr << "Durations.cpp:: The value of MW_ptr->Server_get_operation(Trans_ID,Server_ID) is " << MW_ptr->Server_get_operation(Trans_ID,Server_ID) << std::endl;
   }
    
    
    if((Begin_Status[Server_ID])[Trans_ID]==false)
    {
        /* This is a Begin */
        new_quantile = true;
        duration = Sample_Dist(Trans_ID, Server_ID, 1, Time_Since_Last_Event, new_quantile, current_time, my_reactivate_boolean);
        
        //if(Trans_ID==5)
        if(GV_debug_flag==23)    
        {
            std::cerr << "Durations.cpp:: The sojourn-time for the \"Begin\" of a new transaction on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , is being sampled !!!!"<< std::endl;
        }

        print_trace_method(print_trace, Server_ID, Trans_ID, current_time, std::string("Begin"), 0);
        
    }else if((MW_ptr->Server_get_operation(Trans_ID,Server_ID)).compare(string("Commit"))==0)
    {
        /* This is a Commit */
        new_quantile = false;
        duration = Sample_Dist(Trans_ID, Server_ID, 2, Time_Since_Last_Event, new_quantile, current_time, my_reactivate_boolean);
        
        //if(Trans_ID==5)
        if(GV_debug_flag==23)    
        {
            std::cerr << "Durations.cpp:: The sojourn-time for the \"Commit\" of a transaction on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , is being sampled !!!!"<< std::endl;
        }
        
        print_trace_method(print_trace, Server_ID, Trans_ID, current_time, std::string("Commit"), /*(Server_oper_index[Trans_ID])[Server_ID] + 1*/MW_ptr->Server_get_Trans_oper_ID(Trans_ID,Server_ID));        
        
    }else if((MW_ptr->Server_get_operation(Trans_ID,Server_ID)).compare(string("Abort"))==0)
    {
        /* This is an Abort */
        new_quantile = false;
        duration = Sample_Dist(Trans_ID, Server_ID, 3, Time_Since_Last_Event, new_quantile, current_time, my_reactivate_boolean);
        
        //if(Trans_ID==5)
        if(GV_debug_flag==23)    
        {
            std::cerr << "Durations.cpp:: The sojourn-time of " << duration << " at time " << current_time << " for the \"Abort\" of a transaction on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , is being sampled !!!!"<< std::endl;
        }
        
        print_trace_method(print_trace, Server_ID, Trans_ID, current_time, std::string("Abort"), /*(Server_oper_index[Trans_ID])[Server_ID] + 1*/MW_ptr->Server_get_Trans_oper_ID(Trans_ID,Server_ID));        
        
    }else
    {
        
        //if(ClientObjects[Trans_ID - 1]->get_Trans_Profile_oper_ID() > 0)
        if(MW_ptr->Server_get_Trans_oper_ID(Trans_ID,Server_ID) > 0)
        {
            /* This is a normal operation  */
            new_quantile = false;
            duration = Sample_Dist(Trans_ID, Server_ID, 0, Time_Since_Last_Event, new_quantile, current_time, my_reactivate_boolean);
            
            
            //if(Trans_ID==5)
            if(GV_debug_flag==23)    
            {
                std::cerr << "Durations.cpp:: With a duration value of "<< duration <<", The sojourn-time for operation " << MW_ptr->Server_get_operation(Trans_ID,Server_ID) << " of a new transaction on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , has been sampled !!!!"<< std::endl;
            }
            
            print_trace_method(print_trace, Server_ID, Trans_ID, current_time, MW_ptr->Server_get_operation(Trans_ID,Server_ID), /*(Server_oper_index[Trans_ID])[Server_ID] + 1*/MW_ptr->Server_get_Trans_oper_ID(Trans_ID,Server_ID));            

        }else
        {
            //if(Trans_ID==5)
            if(GV_debug_flag==23)    
            {
                std::cerr << "Durations.cpp:: I DONT KNOW WHY THIS IS BEING CALLED:: The sojourn-time for operation " << ((Oper_Data_for_Replicas[Trans_ID])[(Oper_Data_for_Replicas[Trans_ID]).size()])[0] << " of a new transaction on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , is being sampled !!!!"<< std::endl;
            }            
            
            /* Sample according to exponential distribution */
            duration = SojournTimeValue();

            bool Threshold_Reached = false;

            /* If the system has begun to struggle significantly with each new operation then set the boolean to indicate this.
             * Note: the Current_Num_of_Oper[Server_ID] data structure has already been updated in Mobius if this reactivation 
             * has been triggered by an operation coming to an end   */
            if((Server_ID==1)&&(GV_Efficient_Oper_Capacity_Server_1 <= Current_Num_of_Oper[Server_ID]))
            {
                Threshold_Reached = true;
            }


            if((Server_ID==2)&&(GV_Efficient_Oper_Capacity_Server_2 <= Current_Num_of_Oper[Server_ID]))
            {
                Threshold_Reached = true;
            }

            if(Threshold_Reached)
            {
                /* Both exponential and linear scaling */
                duration = std::pow( static_cast<double>(2 + Current_Num_of_Oper[Server_ID] - ((Server_ID==1)?GV_Efficient_Oper_Capacity_Server_1:GV_Efficient_Oper_Capacity_Server_2)),GV_ReacModel_PolyPower)*duration;

            }else
            {
                /* No linear scaling due to efficiency threshold being reached */
                duration = duration;
            }                
            if(GV_debug_flag==23)
            {
                std::cerr << "Durations.cpp:: An exponentially distributed value of " << duration <<" , due to a transaction operation ending on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , is being returned "<< std::endl;
            }                                            
        }

    }
        
    /* Store the value of the computed Sojourn-time */    
    (Sampled_SojournTimes[Server_ID])[Trans_ID] = duration;
         
    /* Store the current time as the time of the last event */
    (PreviousEventTime[Server_ID])[Trans_ID] = current_time;
    
    return duration;
}


void print_trace_method(bool print_trace, int Server_ID, int Trans_ID, double current_time, std::string operation, int Operation_ID)
{
    /* Print trace data if this is the beginning of an operation */
    if((print_trace==true)&&( (  (GV_Enable_Server_1_View_Trace==1)&&(Server_ID==1)  )    || (  (GV_Enable_Server_2_View_Trace==1)&&(Server_ID==2)  )  ))
    {
        /***** Send Trace data ******/
        /* Format:: Location : Location_ID : Origin : Destination : Trans_ID : Trans_Type : Operation_ID : Operation-type : Operation result : Time-stamp */

        std::stringstream trace_data_stream;

        trace_data_stream << std::string("Durations_cpp Server_") << Server_ID << std::string(" ") << Trans_ID << std::string(" Client ") << std::string(" Server_") << Server_ID << std::string(" ") << Unbounded_Trans_ID[Trans_ID] << std::string(" ")  << (ClientObjects[Trans_ID - 1])->get_Trans_type_names((ClientObjects[Trans_ID - 1])->get_Trans_Profile_ID()) << std::string(" ") << Operation_ID << std::string(" ") << operation << std::string(" ") << std::string("NULL") << std::string(" ") << current_time << std::string("\n");

        if(GV_debug_flag==23)
        {
           cerr << trace_data_stream.str() << endl;
        }

        /* Don't send trace data if the number of Simulation runs has exceeded the intended number */
        if(!(GV_Experiment_Runs < Experiment_Runs_Count))
        {
            /* Record the number of times this activity fires (as a "sanity" check for the trace logger) */
            ActionFiringCounts[(Server_ID==1)? 2:4] = ActionFiringCounts[(Server_ID==1)? 2:4 ] + 1;
            Trace_data_Client->send_data(trace_data_stream.str()); 
        }


        /****************************/        
    }    
};


double Sample_Dist(int Trans_ID, int Server_ID, int Oper_case, double Time_Since_Last_Event, bool new_quantile, double current_time, bool my_reactivate_boolean)
{
    double dur = 0.0;
    
    bool reset_previous_quantile_flag = false; /* This flag, when true, indicates to the sojourn-time sampling method that 
                                                * the value of the previous quantile should be set to zero.
                                                */
    
    int flat_index = 0;
    
    /* Durations are sampled according to empirical distributions */
    if((MW_ptr->Server_get_Trans_oper_ID(Trans_ID,Server_ID) > 0)&&(Oper_case==0))
    {
        int index_1 = (ClientObjects[Trans_ID - 1])->get_Trans_Profile_ID();
        int index_2 = MW_ptr->Server_get_Trans_oper_ID(Trans_ID,Server_ID);
        flat_index = (ClientObjects[Trans_ID - 1])->get_flat_Trans_Oper_index(index_1,index_2);       
    }   
    
    int index_3 = 0;
    
    if(Oper_case==1)
    {
        /* Begin */
        index_3 = (Emp_Distributions[Server_ID]).size() - 2;
        
    }else if(Oper_case==2)
    {
        /* Commit */
        /* Set flag to indicate to duration sampling method that the last quantile stored
         * should be reset to zero 
         */
        reset_previous_quantile_flag = true;
        
        index_3 = (Emp_Distributions[Server_ID]).size() - 1;
        
    }else if(Oper_case==3)
    {
        /* Abort */
        index_3 = (Emp_Distributions[Server_ID]).size();
        
    }else
    {
        /* "non-Edge" Transaction Operation */
        index_3 = flat_index;
    }
    
    dur = sample(&((Emp_Distributions[Server_ID])[index_3]), new_quantile, reset_previous_quantile_flag, Trans_ID, Server_ID, current_time, my_reactivate_boolean);

    
    //if(Trans_ID==5)
    if(GV_debug_flag==23)
    {
        std::cerr << "Durations.cpp:: The initial sampled value for the operation's sojourn-time using the index " << index_3 << " is " << dur <<" , for a transaction operation on Server " << Server_ID << " , on behalf of Client " << Trans_ID << std::endl;
    }
    
    dur = Residual_Sojourn_Time(dur, Server_ID, Trans_ID, Time_Since_Last_Event);
            
    //if(Trans_ID==5)
    if(GV_debug_flag==23)    
    {
            std::cerr << "Durations.cpp:: A \"scaled\" empirically distributed value of " << dur <<" on Server " << Server_ID << " , on behalf of Client " << Trans_ID << " , is being returned "<< std::endl;            
    }
    
        /* For the benefit of the mechanism that is Peter's preferred way of computing residuals,
     * store the estimated duration for this operation 
     */
    
    if(GV_DefaultResidualTimes)
    {
        /* If this is a new sojourn-time then return the entire duration */
        if(!my_reactivate_boolean)
        {
            (previous_sampled_sojtimes[Trans_ID-1])[Server_ID-1] = dur + current_time;            
        }else
        {
            /* Otherwise, return the residual */
            dur = (previous_sampled_sojtimes[Trans_ID-1])[Server_ID-1] - current_time;
        }        
    }
    
    return dur;
};

double Residual_Sojourn_Time(double rv, int Server_ID, int Trans_ID, double TimeSinceLastEvent)
{
    /* This method scales the randomly generated sojourn-time to account for the fact
     * that the operation may have already performed some operations.
     * 
     * The scaling law is such that, 1) for a given number of concurrent operations, it 
     * is an exponentially decreasing scaling, and 2) for a given amount of already consumed
     * sojourn-time, it is a polynomially increasing scaling.
     * 
     */
    
    double residual_time = 0;
        
    /* Is this a reactivation? If not, then the "residual time" exponent should be zero, indicating that this is the first time
     * this operation is having its Sojourn-time computed. Consequently, there should be no dependence on time-already-consumed
     * for this operation in that eventuality; only a dependence on the number of currently executing operations.   
     */
        
     //if(((Sampled_SojournTimes[Server_ID])[Trans_ID]!=0)||((Sampled_SojournTimes[Server_ID])[Trans_ID] != TimeSinceLastEvent))
     if( (!areEqualRel((Sampled_SojournTimes[Server_ID])[Trans_ID], 0, std::numeric_limits<double>::epsilon())) || (!areEqualRel((Sampled_SojournTimes[Server_ID])[Trans_ID], TimeSinceLastEvent, std::numeric_limits<double>::epsilon())) )
     {
         /* this is a reactivation */
        
        //if(GV_debug_flag==23)
        if(GV_debug_flag==23)
        {
             std::cerr << "Durations.cpp:: The value of Time_Since_Last_Event is " << TimeSinceLastEvent << " , on Server " << Server_ID << " , on behalf of Client " << Trans_ID << "!!!!"<< std::endl;
             std::cerr << "Durations.cpp:: The value of (Sampled_SojournTimes[Server_ID])[Trans_ID] is " << (Sampled_SojournTimes[Server_ID])[Trans_ID] << " , on Server " << Server_ID << " , on behalf of Client " << Trans_ID << "!!!!"<< std::endl;             
        }
         
         residual_time = ((Sampled_SojournTimes[Server_ID])[Trans_ID])/((Sampled_SojournTimes[Server_ID])[Trans_ID] - TimeSinceLastEvent);
     }
    
     //if(GV_debug_flag==23)
     //if(Trans_ID==5)
     if(GV_debug_flag==23)
     {
         std::cerr << "Durations.cpp:: The value of residual_time is " << residual_time << " , on Server " << Server_ID << " , on behalf of Client " << Trans_ID << "!!!!"<< std::endl;
     }        
    
    bool Threshold_Reached = false;
    
    /* If the system has begun to struggle significantly with each new operation then set the boolean to indicate this.
     * Note: the Current_Num_of_Oper[Server_ID] data structure has already been updated in Mobius if this reactivation 
     * has been triggered by an operation coming to an end   */
    if((Server_ID==1)&&(GV_Efficient_Oper_Capacity_Server_1 <= Current_Num_of_Oper[Server_ID]))
    {
        Threshold_Reached = true;
    }
    
    
    if((Server_ID==2)&&(GV_Efficient_Oper_Capacity_Server_2 <= Current_Num_of_Oper[Server_ID]))
    {
        Threshold_Reached = true;
    }
     
    if(Threshold_Reached)
    {
        /* Both exponential and polynomial scaling */
        double scaling = exp(-1.0*(GV_ReacModel_ExpScaling*residual_time));
        
        if(scaling>GV_ExpScaling_LowerBound)
        {
            rv = scaling*std::pow( static_cast<double>(2 + Current_Num_of_Oper[Server_ID] - ((Server_ID==1)?GV_Efficient_Oper_Capacity_Server_1:GV_Efficient_Oper_Capacity_Server_2)),GV_ReacModel_PolyPower)*rv;            
        
        }else
        {
            rv = GV_ExpScaling_LowerBound*std::pow( static_cast<double>(2 + Current_Num_of_Oper[Server_ID] - ((Server_ID==1)?GV_Efficient_Oper_Capacity_Server_1:GV_Efficient_Oper_Capacity_Server_2)),GV_ReacModel_PolyPower)*rv;            
        }
        
    }else
    {
        /* Both exponential and polynomial scaling */
        double scaling = exp(-1.0*(GV_ReacModel_ExpScaling*residual_time));
        
        if(scaling>GV_ExpScaling_LowerBound)
        {
            /* No polynomial scaling (of the generated sojourn-time) due to efficiency threshold being reached */
            rv = scaling*rv;            
        
        }else
        {
            rv = GV_ExpScaling_LowerBound*rv;
        }
    }
    
    return rv;
}

bool areEqualRel(double a, double b, double epsilon) 
{
    //FLT_EPSILON;
    //std::numeric_limits<double>::epsilon()
    
    return (std::abs(a - b) <= std::abs(std::min(a,b))*epsilon);
}
