#include "DIDERO_PC.h"
#include <vector>
#include "Client_Interface.h"
#include "Server_Interface.h"
#include "Server_Interface_2.h"
#include "Middleware_Interface.h"
#include "string"
#include <sstream>

extern short GV_debug_flag;
extern vector<Client_Interface*> ClientObjects;        /* Array of pointers to Client Object interface*/
extern Server_Interface* Server_1;        /* Pointer to Server interface*/
extern Server_Interface* Server_2;        /* Pointer to Server interface*/
extern Middleware_Interface* MW_ptr;      /* Pointer to Middleware Interface */      
extern vector<bool> Exec_Oper_on_both_Servers;         /* Array of booleans to indicate that an operation 
                                                 * has been carried out on both servers. 
                                                 */
extern vector<string> Oper_results;             /* Array of strings to store the result of the faster server */
extern map<int, map< int, map< int , string> > > Trans_Oper_Replica_Results;       /* Stores the results of operations executed by the 
Client_ID  Operation_ID   Server_ID  Result                                           replicas. These results are compared by the Middlewware 
                                                                                    * for possible inconsistencies. 
                                                                                    */

extern map<int, map<int, vector<string> > > Oper_Data_for_Replicas;   /*Array to store the data that defines each operation in a Transaction */
/* Client_ID Operation_ID   Data */

extern map< int, Place* > Unblocked_place_ptrs;        /* Pointers to places that unblock blocked Transitions */
extern map< int, Place* > Blocked_place_ptrs;        /* Pointers to places that indicate the blocked status of each Transitions */

/* Data structure that keeps count of which operation is currently being executed by a given replica, with respect to a given client */
extern map<int, map<int, int>  > Server_oper_index;

string Exec_Trans_Method(int ID, int Server_ID)
{
    static string result = string("");
    static string Blocked("Blocked");
    static string Wait("Wait for Blocked Transaction");
    static string disagree("Servers Disagree");
    static string commit_success("Commit success");
    static string abort("Abort");    
    
    /* Read shared data and execute the corresponding transaction */
    if(Server_ID==1)
    {           
        if(GV_debug_flag==26)
        {
            //MW_ptr->Server_get_operation(ID, Server_ID), MW_ptr->Server_get_operation_data(ID, Server_ID)
            //if( (ClientObjects[ID-1]->get_Shared_Data_Operation()).compare(std::string("Abort"))==0)
            if((MW_ptr->Server_get_operation(ID, Server_ID)).compare(std::string("Abort"))==0)
            {
                cerr << "Exec_Trans_Method.cpp:: Transaction for Client " << ID << " is about to experience an abort on Server 1." << endl;
            }
        }
            
        /* Execute requested operation and set the operation string to indicate whether the operation was
         * indeed successful.
         */

        result.clear();
        //result = Server_1->Operation(ID, ClientObjects[ID-1]->get_Shared_Data_Operation(), ClientObjects[ID-1]->get_Shared_Data_Key());
        result = Server_1->Operation(ID, MW_ptr->Server_get_operation(ID, Server_ID), MW_ptr->Server_get_operation_data(ID, Server_ID));
         
        /* If the transaction was blocked then do not store any evidence of this method being called, as it might be 
           called again when the transaction is unblocked by the server.
         */
         
         if(result.compare(std::string("Blocked"))==0)
         {
            if(GV_debug_flag==26)
            {
                cerr << "Exec_trans_Method.cpp:: Exec_Trans_Method.cpp: result on Server 1 for Client " << ID << " is " << result << " and Shared_Data_Operation is " << result << endl;
            }             
             
         }else
         {
            /* set the result of the operation */
            MW_ptr->Server_set_operation_result(ID, Server_ID, result);
             
            if(GV_debug_flag==20)
            {
                cerr << "Exec_trans_Method.cpp:: Exec_Trans_Method.cpp: result on Server 1 for Client "<< ID << " is " << result << " and Shared_Data_Operation is " << MW_ptr->Server_get_operation(ID, Server_ID) << endl;
            }
             
            if((GV_debug_flag==26)&&(result.compare(std::string("Abort"))==0))
            {
                cerr << "Exec_trans_Method.cpp:: Exec_Trans_Method.cpp: result on Server 1 for Client "<< ID << " is " << result << " and Shared_Data_Operation is " << MW_ptr->Server_get_operation(ID, Server_ID) << endl;
            }                          
         }
    }
    
    if(Server_ID==2)
    {   
        if(GV_debug_flag==26)
        {
            if((MW_ptr->Server_get_operation(ID, Server_ID)).compare(std::string("Abort"))==0)
            {
                cerr << "Exec_Trans_Method.cpp:: Transaction object " << ID << " is about to experience an abort on Server 2." << endl;
            }
        }
            
        /* Execute requested operation and set the operation string to indicate whether the operation was
         * indeed successful.
         */

        if(GV_debug_flag==20)
        {
            cerr << "Exec_trans_Method.cpp:: The value of MW_ptr->Server_get_operation(ID, Server_ID) is " << MW_ptr->Server_get_operation_data(ID, Server_ID) << endl;
            cerr << "Exec_trans_Method.cpp:: The value of MW_ptr->Server_get_operation_data(ID, Server_ID) is " << MW_ptr->Server_get_operation_data(ID, Server_ID) << endl;
        }
        
        result.clear(); 
        result = Server_2->Operation(ID, MW_ptr->Server_get_operation(ID, Server_ID), MW_ptr->Server_get_operation_data(ID, Server_ID));
         
        /* set the result of the operation */
        MW_ptr->Server_set_operation_result(ID, Server_ID, result);
        
        if((GV_debug_flag==26)&&(result.compare(std::string("Abort"))==0))
        {
            cerr << "Exec_trans_Method.cpp:: Exec_Trans_Method.cpp: result on Server 2 for Client " << ID << " is " << result << " and Shared_Data_Operation is " << MW_ptr->Server_get_operation(ID, Server_ID) << endl;
        }
    }
   
    
    return result; 
}
