#include "DIDERO_PC.h"
#include <map>
//#include <iomanip>
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>

using namespace std;

map< int, map<int, map<double, double> > > Emp_Distributions;

std::string SplitFilename(const std::string& str);

int Init_Emp_Dists(string PathToFileWithDataLocations)
{
    int localbool = 0;
 
    /* Set up stringstreams and strings for determiing and appending the location of this simulation exe
     * to the beginning of each distribution file's relative path
     *  
     */
    
            char buffer[MAX_PATH];//always use MAX_PATH for filepaths
            GetModuleFileName(NULL,buffer,sizeof(buffer));
            //std::cerr << "Filepath:" << buffer << "\n";

            stringstream temp_ss;
            stringstream temp_processpath;
            
            /* Store path to process*/
            /* Make sure to search for last backslash and ignore all string data after this */
            temp_processpath.str(std::string());
            temp_processpath << buffer;
            std::string processpath = SplitFilename(temp_processpath.str()); 
            std::replace(processpath.begin(), processpath.end(),'\\','/');
    
            if(GV_debug_flag==23)
            {
                std::cerr << PathToFileWithDataLocations <<endl;                
            }

    int Number_of_Dists_per_Server = 37;
    
     /*********************** Read in the Empirical Distributions and initialize Data structures ******************************/ 
     
     Emp_Distributions.clear();
     
     vector<double> unsorted_dist;
     unsorted_dist.clear();
          
     int dist_counter = 0;
     double Prob_percentile = 0;
     double soj_time = 0;
     
    /************ Read in the input files containing distributional information into appropriate data structures **********/
    ifstream fileIn_EmpDist_Paths(PathToFileWithDataLocations.c_str());   //Stream that enables access to the file containing the paths to the Empirical distributions file
    
    /* Has the file stream been initialised successfully? */

    if(!fileIn_EmpDist_Paths)
    {
        if(GV_debug_flag==23)
        {
            cerr << "Init_Emp_Dists.cpp:   Can't open input file " << PathToFileWithDataLocations.c_str() << ".\n" << endl;
        }
        
        localbool = 0;
        return localbool;
    }
    else
    {
	if(GV_debug_flag==23)
        {
            cerr << "Init_Emp_Dists.cpp: Input file " << PathToFileWithDataLocations.c_str() << " was opened.\n" << endl;
        }
    }    

    if(fileIn_EmpDist_Paths.eof())
    {
        if(GV_debug_flag==23)
        {
            cerr << "Init_Emp_Dists.cpp: " << PathToFileWithDataLocations.c_str() <<" is not formatted correctly. It has no data within it."<< endl;
        }
        
        localbool = 0;                  /* Set a boolean that will indicate to the client that the simulation should terminate.*/
        return localbool;
    }else
    {
        /* Iterate through the file, each time obtaining the path to the Empirical distribution file.
         * Populate a data structure to hold each Empirical distribution. */
        
        string line("");

        /* IDs for the current Server and Operation */
        int Server_ID = 0;
        int Oper_ID = 0;
        
        ifstream fileIn_EmpDist;   //Stream that enables access to the file containing the paths to the Empirical distributions file                            

        
        
        /*Read first line in the file. I assume this should be a line with the string "Section" in it*/
        line.clear();
        getline(fileIn_EmpDist_Paths,line);
        
        while(!fileIn_EmpDist_Paths.eof())
        {
            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << line.c_str();
            line.clear();
            line = std::string("");
            temp_ss >> line;        
        
            fileIn_EmpDist.open(line.c_str());        
            
            /* Initialize variables */
            dist_counter = dist_counter + 1;
            Prob_percentile = 0;
            soj_time = 0;

            /* Construct the IDs for the current Server and Operation pair that determines the current Empirical
             * distribution being read in.
             *  
             */
            
            Server_ID = 1 + (static_cast<int>(dist_counter)/static_cast<int>(Number_of_Dists_per_Server));
            Oper_ID = ((dist_counter)%Number_of_Dists_per_Server)==0?Number_of_Dists_per_Server:((dist_counter)%Number_of_Dists_per_Server) ;            
            
            /* "Clear" the data structure that will hold the Empirical distribution */
            ((Emp_Distributions[Server_ID])[Oper_ID]).clear();
                        
            if(!fileIn_EmpDist)
            {
                if(GV_debug_flag==23)
                {
                    cerr << "Init_Emp_Dists.cpp:   Can't open input file " << line.c_str() << ".\n" << endl;
                }
                
                localbool = 0;
                return localbool;
            }else
            {
                if(GV_debug_flag==23)
                {
                    cerr << "Init_Emp_Dists.cpp: Input file " << line.c_str() << " was opened.\n" << endl;
                }
            }                
            
            string local_line("");
            local_line.clear();

            while(getline(fileIn_EmpDist,local_line))
            {
                if(!local_line.empty())
                {
                    double Emp_prob_value = 0.0;
                    double RandomVariable = 0.0;
                    
                    /* Create a new istringstream each time. Alternative would be to re-initialize a single stream object
                     * which, frankly, is not necessary for my purposes
                     */
                    istringstream iss(local_line);
                    iss >> Emp_prob_value >> RandomVariable;
                    
                    /* Deprecated : The Empirical distributions data contains random variable realisations of value "0": this is not suitable for simulation.
                     * Consequently, when reading in the data, the "0" values are changed to "1" and the other values are scaled by 100.
                     * Note (20141112) : Since Vlad has implemented a more granular time-stamping method there is no more need for the 
                     * distributional data to be scaled.
                     */
                    
                    //((Emp_Distributions[Server_ID])[Oper_ID])[Emp_prob_value] = (RandomVariable==0)?1:(GV_Oper_Dur_Raw_Data_Scaling*RandomVariable);
                    ((Emp_Distributions[Server_ID])[Oper_ID])[Emp_prob_value] = GV_Oper_Dur_Raw_Data_Scaling*RandomVariable;
                }
            }
                        
            if(GV_debug_flag==23)
            {
                cerr << "Init_Emp_Dists.cpp: The "<< dist_counter << "th distribution has been created!!!"<< endl;
            }
            
            /* Reset filestream for next file */
            fileIn_EmpDist.clear();
            fileIn_EmpDist.close();

            line.clear();
            getline(fileIn_EmpDist_Paths,line);            
        }
        
        fileIn_EmpDist_Paths.clear();
        fileIn_EmpDist_Paths.close();
    }
    return localbool;
}
