
//#define WIN32_LEAN_AND_MEAN

//#include <windows.h>

#include "DIDERO_PC.h"

/* KS Comment: This source, when compiled, needs to be linked
 * with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
 * 
 */

extern char DIDERO_PC_Trace_Paths[4*MAX_PATH];

#define DEFAULT_BUFLEN 1024
#define DEFAULT_PORT "28000"

Inter_Proc_Comm_Client::Inter_Proc_Comm_Client(std::string Server_path)
{
    /* Spawn Independent Server Process to listen for Simulation Trace data */
    
    /* The following WIN structs are useful for interacting with the spawned process. They are 
     * not important for our needs but, I suspect, are necessary since the CreateProcess method
     * expects these structures 
     *
     */
    
    STARTUPINFO info={sizeof(info)};
    PROCESS_INFORMATION processInfo;
    
    ////std::cerr << "The path DIDERO_PC_Trace_Paths is " << DIDERO_PC_Trace_Paths << endl;
    
    if (!CreateProcess(Server_path.c_str(), DIDERO_PC_Trace_Paths, NULL, NULL, TRUE, 0, NULL, NULL, &info, &processInfo))
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: The Server process for recording Trace data was not successfully spawned !!!! \n\n";
    
    }else
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: The Server process for recording Trace data WAS successfully spawned !!!! \n\n"; 
    }
    
    ConnectSocket = INVALID_SOCKET;
    result = NULL; 
    ptr = NULL;
    sendbuf = std::string("This is a test\n");
    
    // Initialize Winsock
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if(iResult != 0)
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: WSAStartup failed with error: " << iResult << "\n";
    }else
    {
        ZeroMemory( &hints, sizeof(hints) );
        hints.ai_family = AF_INET;
        hints.ai_socktype = SOCK_STREAM;
        hints.ai_protocol = IPPROTO_TCP;

        /**************** Obtain the Local-Machine's IP address **************/
        char hostname[80];
        if(gethostname(hostname, sizeof(hostname)) == SOCKET_ERROR) 
        {
            std::cerr << "Inter_Proc_Comm_Client.cpp:: Error " << WSAGetLastError() << " when getting local host name." << std::endl;
        }else
        {
            if(GV_debug_flag==5)
            {
                std::cerr << "Inter_Proc_Comm_Client.cpp:: Host name is " << hostname << "." << std::endl;        
            }

            struct hostent *phe = gethostbyname(hostname);
            if(phe == 0)
            {
                std::cerr << "Inter_Proc_Comm_Client.cpp:: Oow! Bad host lookup." << std::endl;
            }else
            {
                /* Copy the IP address from the "hostent" object pointed to by "phe" */
                struct in_addr addr;
                memcpy(&addr, phe->h_addr_list[0], sizeof(struct in_addr));

                if(GV_debug_flag==5)
                {
                    std::cerr << "Inter_Proc_Comm_Client.cpp:: Address : " << inet_ntoa(addr) << std::endl;
                }

                iResult = getaddrinfo(inet_ntoa(addr), DEFAULT_PORT, &hints, &result);
                /******************* IP address Obtained and used to construct struct with socket relevant info ****************/

                if(iResult!=0)
                {
                    std::cerr << "Inter_Proc_Comm_Client.cpp:: getaddrinfo failed with error: " << iResult << "\n";
                    WSACleanup();
                }else
                {   
                    // Attempt to connect to an address until one succeeds
                    for(ptr=result; ptr != NULL ;ptr=ptr->ai_next)
                    {
                        // Create a SOCKET for connecting to server
                        ConnectSocket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
                        if(ConnectSocket==INVALID_SOCKET)
                        {
                            std::cout << "Inter_Proc_Comm_Client.cpp:: socket failed with error: " << WSAGetLastError() << "\n";
                            //KS comment: the following method call terminates the use of the WS2_32.dll
                            WSACleanup();
                        }else
                        {
                            // Connect to server.
                            iResult = connect( ConnectSocket, ptr->ai_addr, static_cast<int>(ptr->ai_addrlen));
                            if(iResult == SOCKET_ERROR)
                            {
                                closesocket(ConnectSocket);
                                ConnectSocket = INVALID_SOCKET;
                            }                            
                        }
                    }

                    //KS comment: did the loop actually result in a ConnectSocket Socket?
                    if(ConnectSocket == INVALID_SOCKET) 
                    {
                        std::cerr << "Inter_Proc_Comm_Client.cpp:: Unable to connect to server!\n";
                        WSACleanup();
                    }
                }               
            }             
        }       
    }
}

int Inter_Proc_Comm_Client::send_data(std::string data)
{
    // Send an initial buffer
    iResult = send( ConnectSocket, data.c_str() , static_cast<int>(strlen(data.c_str())), 0 );
    //iResult = send( ConnectSocket, data->c_str() , static_cast<int>(sizeof(data->c_str())), 0 );
    if(iResult == SOCKET_ERROR) 
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: send failed with error: " << WSAGetLastError() << "\n";
        closesocket(ConnectSocket);
        WSACleanup();
        return 0;
    }else
    {
        if(GV_debug_flag==13)
        {
            std::cerr << "Inter_Proc_Comm_Client.cpp:: Successfully sent Trace data to Server!!!!!" << std::endl;            
        }
    }
        
    return 1;
}

Inter_Proc_Comm_Client::~Inter_Proc_Comm_Client()
{
    std::stringstream local_temp_string;
    local_temp_string.str("");
    local_temp_string.clear();
    
    /* Send Simulation summary data to the Trace logger for each of the Trace views */
    for(int iter = 0; iter < 6; iter++)
    {
        /* Construct summary data string and send it to the Trace logger */
        //ActionFiringCounts.end()
        
        switch(iter)
        {
            case 0:
                local_temp_string << std::string("Simulation Information Client Origin ") << ActionFiringCounts[iter] << "\n";
                break;
            case 1:
                local_temp_string << std::string("Simulation Information Client Result ") << ActionFiringCounts[iter] << "\n";
                break;                
            case 2:
                local_temp_string << std::string("Simulation Information Server_1 Origin ") << ActionFiringCounts[iter] << "\n";
                break;                                
            case 3:
                local_temp_string << std::string("Simulation Information Server_1 Result ") << ActionFiringCounts[iter] << "\n";
                break;
            case 4:
                local_temp_string << std::string("Simulation Information Server_2 Origin ") << ActionFiringCounts[iter] << "\n";
                break;                                                
            case 5:
                local_temp_string << std::string("Simulation Information Server_2 Result ") << ActionFiringCounts[iter] << "\n";
                break;
            default:
                std::cerr << "Initer_Proc_Comm_Client.cpp:: No Simulation summary data was saved." << std::endl;
        }
    }
    
    /* Send the summary data */
    iResult = send( ConnectSocket, (local_temp_string.str()).c_str() , static_cast<int>(strlen((local_temp_string.str()).c_str())), 0 );
    if(iResult == SOCKET_ERROR) 
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: send failed with error: " << WSAGetLastError() << "\n";
        closesocket(ConnectSocket);
        WSACleanup();
    }else
    {
        if(GV_debug_flag==13)
        {
            std::cerr << "Inter_Proc_Comm_Client.cpp:: Successfully sent Trace data to Server!!!!!" << std::endl;            
        }
    }    
    
    // shutdown the connection since no more data will be sent
    iResult = shutdown(ConnectSocket, SD_SEND);
    if(iResult == SOCKET_ERROR) 
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: shutdown failed with error: " << WSAGetLastError() << "\n";
        closesocket(ConnectSocket);
        WSACleanup();
    }else
    {
        // cleanup
        //KS comment: the object pointed to by result, which holds the address-info relevant for this application, is no longer required
        freeaddrinfo(result);
        closesocket(ConnectSocket);
        WSACleanup();            
    }
}

int Inter_Proc_Comm_Client::send_test_data()
{
    // Send an initial buffer
    iResult = send( ConnectSocket, sendbuf.c_str() , static_cast<int>(strlen(sendbuf.c_str())), 0 );

    if(iResult == SOCKET_ERROR) 
    {
        std::cerr << "Inter_Proc_Comm_Client.cpp:: send failed with error: " << WSAGetLastError() << "\n";
        closesocket(ConnectSocket);
        WSACleanup();
        return 0;
    }else
    {
        if(GV_debug_flag==13)
        {
            std::cerr << "Inter_Proc_Comm_Client.cpp:: Successfully sent Trace data to Server!!!!!" << std::endl;            
        }
    }
        
    return 1;    
}
