#include <vector>
#include "Client_Interface.h"
#include "Server_Interface.h"
#include "Server_Interface_2.h"
#include "string"
#include "DIDERO_PC.h"
#include "boost/random.hpp"
#include "boost/generator_iterator.hpp"

/* Initialise and create random number generation function */
typedef boost::mt19937 RNGType;  //Use TYPEDEF so that you only need to change this typedef to change generation procedure
RNGType rng;
boost::uniform_real<double> zero_to_one( 0.0, 1.0 );
boost::variate_generator< RNGType, boost::uniform_real<double> > dice(rng, zero_to_one);

extern vector<Client_Interface*> ClientObjects;        /* Array of pointers to Client Object interface*/
extern Server_Interface* Server_1;        /* Array of pointers to Client Object interface*/
extern Server_Interface* Server_2;        /* Array of pointers to Client Object interface*/
extern vector<bool> Exec_Oper_on_both_Servers;         /* Array of booleans to indicate that an operation 
                                                 * has been carried out on both servers. */
extern short GV_debug_flag;


bool Next_Operation_Method(int ID)
{    
    bool result = true;
    
    static string abort_success("Abort success");
    static std::string commit_success("Commit success");
    
    /* Check to see if a new transaction should be started, given that 
     * a "Commit" or "Abort" was just successfully completed. Otherwise,
     * on to the next operation.
     */
    
    //result = !(((ClientObjects[ID-1]->get_Oper_Exec_Status()).compare(std::string("Abort success"))==0)||((ClientObjects[ID-1]->get_Oper_Exec_Status()).compare(std::string("Commit success"))==0)||((ClientObjects[ID-1]->get_Oper_Exec_Status()).compare(std::string("Blocked"))==0));
    result = !(((ClientObjects[ID-1]->get_Oper_Exec_Status()).compare(abort_success)==0)||((ClientObjects[ID-1]->get_Oper_Exec_Status()).compare(commit_success)==0));
    
    if(GV_debug_flag==5)
    {
        cerr << "Next_Operation.cpp::  \"result\" is " << result << " for client " << ID << endl;
    }
    
    if(result)
    {
        /* Check to see that the client's transaction has not been blocked and
         * the transaction is not scheduled to be aborted. Note: for TPC-C, 1% of
         * Noew-order transactions need to be aborted 
         * 
         */        
        
        bool Neworder_rollback_flag = false;
        
        /* If Neworder Transaction has just begun executing, rollback this transaction instance
         * with probability 1% 
         */
        
        if((((ClientObjects[ID - 1])->get_Trans_type_names((ClientObjects[ID - 1])->get_Trans_Profile_ID())).compare(std::string("New_Order"))==0  )&&(ClientObjects[ID - 1]->get_Trans_Profile_oper_ID() < 1)&&(GV_TPCC))
        {
            /* Signal an Abort for this transaction, with probability 1% */
           if(dice() <= 0.01)
           {
               Neworder_rollback_flag = true;
           }
        }
        
        if(((ClientObjects[ID-1]->get_Oper_Exec_Status()).compare(std::string("Abort"))!=0)&&(!Neworder_rollback_flag))
        {
            string local_result("");
            local_result = ClientObjects[ID-1]->Transaction_Instance();
           
           //if(GV_debug_flag==5)
           if(GV_debug_flag==23)
           {
               cerr << "Next_Operation.cpp:: Client " << ID << " 's next operation: " << local_result << endl;
           }

        }else
        {
            {
                ClientObjects[ID-1]->set_Shared_Data_Operation(std::string("Abort"));
                ClientObjects[ID-1]->set_Shared_Data_Key(std::string(""));
                
                //if(GV_debug_flag==5)
                if(GV_debug_flag==26)                
                {
                    cerr << "Next_Operation.cpp:: Client " << ID << "'s Transaction is to be aborted " << endl;
                }
            }
        }        
    }else
    {
        if(GV_debug_flag==5)
        {
                cerr << "Next_Operation.cpp:: A new transaction is to be started for Client " << ID << endl;
        }        
    }
    
    return result;
}
